@@ -10,14 +10,13 @@ use crossterm::event::{MouseButton, MouseEventKind};
 use smallvec::smallvec;
 
 use crate::app::{Action, MouseArea};
-use crate::command::Command;
 use crate::config::Config;
 use crate::device_kind::DeviceKind;
 use crate::device_widget::DeviceWidget;
 use crate::dropdown_widget::DropdownWidget;
 use crate::node_widget::NodeWidget;
-use crate::object::ObjectId;
 use crate::view::{self, ListKind, VolumeAdjustment};
+use crate::wirehose::ObjectId;
 
 /// ObjectList stores information for filtering and displaying a subset of
 /// objects from a [`View`](`crate::view::View`).
@@ -34,7 +33,7 @@ pub struct ObjectList {
     /// Default device type to use for defaults and node rendering
     device_kind: Option<DeviceKind>,
     /// Target dropdown state
-    pub list_state: ListState,
+    pub dropdown_state: ListState,
     /// Targets
     pub targets: Vec<(view::Target, String)>,
 }
@@ -51,23 +50,23 @@ impl ObjectList {
     }
 
     pub fn down(&mut self, view: &view::View) {
-        if self.list_state.selected().is_some() {
-            self.list_state.select_next();
+        if self.dropdown_state.selected().is_some() {
+            self.dropdown_state.select_next();
         } else {
             let new_selected = view.next_id(self.list_kind, self.selected);
             if new_selected.is_some() {
-                self.selected = new_selected;
+                self.select(new_selected);
             }
         }
     }
 
     pub fn up(&mut self, view: &view::View) {
-        if self.list_state.selected().is_some() {
-            self.list_state.select_previous();
+        if self.dropdown_state.selected().is_some() {
+            self.dropdown_state.select_previous();
         } else {
             let new_selected = view.previous_id(self.list_kind, self.selected);
             if new_selected.is_some() {
-                self.selected = new_selected;
+                self.select(new_selected);
             }
         }
     }
@@ -84,106 +83,101 @@ impl ObjectList {
         if let Some((targets, index)) = targets {
             if !targets.is_empty() {
                 self.targets = targets;
-                self.list_state.select(Some(index));
+                self.dropdown_state.select(Some(index));
             }
         }
     }
 
     fn selected_target(&self) -> Option<&view::Target> {
-        self.list_state
+        self.dropdown_state
             .selected()
             .and_then(|index| self.targets.get(index))
             .map(|(target, _)| target)
     }
 
-    pub fn dropdown_activate(&mut self, view: &view::View) -> Vec<Command> {
+    pub fn dropdown_activate(&mut self, view: &view::View) {
         // Just open the dropdown if it's not showing yet.
-        if self.list_state.selected().is_none() {
+        if self.dropdown_state.selected().is_none() {
             self.dropdown_open(view);
-            return Vec::default();
+            return;
         }
 
-        let commands = self
-            .selected
-            .zip(self.selected_target())
-            .map(|(object_id, &target)| view.set_target(object_id, target))
-            .into_iter()
-            .flatten()
-            .collect();
-        self.list_state.select(None);
-        commands
+        if let (Some(object_id), Some(&target)) =
+            (self.selected, self.selected_target())
+        {
+            view.set_target(object_id, target);
+        };
+
+        self.dropdown_state.select(None);
     }
 
     pub fn dropdown_close(&mut self) {
-        self.list_state.select(None);
+        self.dropdown_state.select(None);
     }
 
-    pub fn set_target(
-        &mut self,
-        view: &view::View,
-        target: view::Target,
-    ) -> Vec<Command> {
-        self.list_state.select(None);
-        self.selected
-            .map(|object_id| view.set_target(object_id, target))
-            .into_iter()
-            .flatten()
-            .collect()
+    pub fn set_target(&mut self, view: &view::View, target: view::Target) {
+        self.dropdown_state.select(None);
+        if let Some(object_id) = self.selected {
+            view.set_target(object_id, target);
+        };
     }
 
-    pub fn toggle_mute(&mut self, view: &view::View) -> Vec<Command> {
+    pub fn toggle_mute(&mut self, view: &view::View) {
         if matches!(self.list_kind, ListKind::Device) {
-            return Vec::new();
+            return;
+        }
+        if let Some(node_id) = self.selected {
+            view.mute(node_id);
         }
-        self.selected
-            .and_then(|node_id| view.mute(node_id))
-            .into_iter()
-            .collect()
     }
 
     pub fn set_absolute_volume(
         &mut self,
         view: &view::View,
         volume: f32,
-    ) -> Vec<Command> {
+        max: Option<f32>,
+    ) -> bool {
         if matches!(self.list_kind, ListKind::Device) {
-            return Vec::new();
+            return false;
         }
-        self.selected
-            .and_then(|node_id| {
-                view.volume(node_id, VolumeAdjustment::Absolute(volume))
-            })
-            .into_iter()
-            .collect()
+        if let Some(node_id) = self.selected {
+            return view.volume(
+                node_id,
+                VolumeAdjustment::Absolute(volume),
+                max,
+            );
+        }
+        false
     }
 
     pub fn set_relative_volume(
         &mut self,
         view: &view::View,
         volume: f32,
-    ) -> Vec<Command> {
+        max: Option<f32>,
+    ) -> bool {
         if matches!(self.list_kind, ListKind::Device) {
-            return Vec::new();
+            return false;
         }
-        self.selected
-            .and_then(|node_id| {
-                view.volume(node_id, VolumeAdjustment::Relative(volume))
-            })
-            .into_iter()
-            .collect()
+        if let Some(node_id) = self.selected {
+            return view.volume(
+                node_id,
+                VolumeAdjustment::Relative(volume),
+                max,
+            );
+        }
+        false
     }
 
-    pub fn set_default(&mut self, view: &view::View) -> Vec<Command> {
+    pub fn set_default(&mut self, view: &view::View) {
         if matches!(self.list_kind, ListKind::Device) {
-            return Vec::new();
+            return;
+        }
+        if let (Some(node_id), Some(device_kind)) =
+            (self.selected, self.device_kind)
+        {
+            view.set_default(node_id, device_kind);
         }
-        self.selected
-            .zip(self.device_kind)
-            .and_then(|(node_id, device_kind)| {
-                view.set_default(node_id, device_kind)
-            })
-            .into_iter()
-            .collect()
     }
 
     fn selected_index(&self, view: &view::View) -> Option<usize> {
@@ -191,11 +185,19 @@ impl ObjectList {
             .and_then(|selected| view.position(self.list_kind, selected))
     }
 
+    fn select(&mut self, object_id: Option<ObjectId>) {
+        self.selected = object_id;
+        // Close the dropdown in case it is open for the previously-selected
+        // object. This can happen when the object is removed from PipeWire
+        // while the dropdown is open.
+        self.dropdown_close();
+    }
+
     /// Reconciles changes to objects, viewport, and selection.
     pub fn update(&mut self, area: Rect, view: &view::View) {
         let selected_index = self.selected_index(view).or_else(|| {
             // There's nothing selected! Select the first item and try again.
-            self.selected = view.next_id(self.list_kind, None);
+            self.select(view.next_id(self.list_kind, None));
             self.selected_index(view)
         });
 
@@ -239,7 +241,7 @@ impl ObjectList {
                         self.top = selected_index;
                     }
                 }
-                None => self.selected = None, // The selected object is gone!
+                None => self.select(None), // The selected object is gone!
             }
         }
     }
@@ -258,9 +260,9 @@ impl ObjectList {
     }
 }
 
-pub struct ObjectListWidget<'a> {
+pub struct ObjectListWidget<'a, 'b> {
     pub object_list: &'a mut ObjectList,
-    pub view: &'a view::View,
+    pub view: &'a view::View<'b>,
     pub config: &'a Config,
 }
 
@@ -270,7 +272,7 @@ struct ObjectListRenderContext<'a> {
     objects_visible: usize,
 }
 
-impl ObjectListWidget<'_> {
+impl ObjectListWidget<'_, '_> {
     fn render_node_list(
         &mut self,
         node_kind: view::NodeKind,
@@ -293,25 +295,25 @@ impl ObjectListWidget<'_> {
             let selected = self
                 .object_list
                 .selected
-                .map(|id| id == object.id)
+                .map(|id| id == object.object_id)
                 .unwrap_or_default();
             NodeWidget::new(
+                self.config,
+                self.object_list.device_kind,
                 object,
                 selected,
-                self.object_list.device_kind,
-                self.config,
             )
             .render(object_area, buf, mouse_areas);
         }
 
         // Show the target dropdown?
-        if self.object_list.list_state.selected().is_some() {
+        if self.object_list.dropdown_state.selected().is_some() {
             // Get the area for the selected object
             if let Some((_, object_area)) =
                 objects_and_areas.iter().find(|(object, _)| {
                     self.object_list
                         .selected
-                        .map(|id| id == object.id)
+                        .map(|id| id == object.object_id)
                         .unwrap_or_default()
                 })
             {
@@ -350,7 +352,7 @@ impl ObjectListWidget<'_> {
             let selected = self
                 .object_list
                 .selected
-                .map(|id| id == object.id)
+                .map(|id| id == object.object_id)
                 .unwrap_or_default();
             DeviceWidget::new(object, selected, self.config).render(
                 object_area,
@@ -360,13 +362,13 @@ impl ObjectListWidget<'_> {
         }
 
         // Show the target dropdown?
-        if self.object_list.list_state.selected().is_some() {
+        if self.object_list.dropdown_state.selected().is_some() {
             // Get the area for the selected object
             if let Some((_, object_area)) =
                 objects_and_areas.iter().find(|(object, _)| {
                     self.object_list
                         .selected
-                        .map(|id| id == object.id)
+                        .map(|id| id == object.object_id)
                         .unwrap_or_default()
                 })
             {
@@ -385,7 +387,7 @@ impl ObjectListWidget<'_> {
     }
 }
 
-impl StatefulWidget for &mut ObjectListWidget<'_> {
+impl StatefulWidget for &mut ObjectListWidget<'_, '_> {
     type State = Vec<MouseArea>;
 
     fn render(self, area: Rect, buf: &mut Buffer, state: &mut Self::State) {
@@ -508,49 +510,61 @@ impl StatefulWidget for &mut ObjectListWidget<'_> {
 #[cfg(test)]
 mod tests {
     use super::*;
-    use crate::capture_manager::CaptureManager;
     use crate::config;
-    use crate::event::MonitorEvent;
-    use crate::media_class::MediaClass;
-    use crate::state::State;
+    use crate::mock;
     use crate::view::{ListKind, NodeKind, View};
+    use crate::wirehose::{state::State, PropertyStore, StateEvent};
 
-    fn init() -> (State, CaptureManager) {
+    fn init() -> (State, mock::WirehoseHandle) {
         let mut state = State::default();
-        let mut capture_manager = CaptureManager::default();
+        let wirehose = mock::WirehoseHandle::default();
 
         for i in 0..10 {
-            let obj_id = ObjectId::from_raw_id(i);
+            let object_id = ObjectId::from_raw_id(i);
+            let mut props = PropertyStore::default();
+            props.set_node_description(String::from("Test node"));
+            props.set_media_class(String::from("Stream/Output/Audio"));
+            props.set_media_name(String::from("Media name"));
+            props.set_node_name(String::from("Node name"));
+            props.set_object_serial(i as u64);
+            let props = props;
+
             let events = vec![
-                MonitorEvent::NodeDescription(
-                    obj_id,
-                    String::from("Test node"),
-                ),
-                MonitorEvent::NodeMediaClass(
-                    obj_id,
-                    MediaClass::from("Stream/Output/Audio"),
-                ),
-                MonitorEvent::NodeMediaName(obj_id, String::from("Media name")),
-                MonitorEvent::NodeName(obj_id, String::from("Node name")),
-                MonitorEvent::NodeObjectSerial(obj_id, i as i32),
-                MonitorEvent::NodePeaks(obj_id, vec![0.0, 0.0], 512),
-                MonitorEvent::NodePositions(obj_id, vec![0, 1]),
-                MonitorEvent::NodeRate(obj_id, 44100),
-                MonitorEvent::NodeVolumes(obj_id, vec![0.0, 0.0]),
-                MonitorEvent::NodeMute(obj_id, false),
+                StateEvent::NodeProperties { object_id, props },
+                StateEvent::NodePeaks {
+                    object_id,
+                    peaks: vec![0.0, 0.0],
+                    samples: 512,
+                },
+                StateEvent::NodePositions {
+                    object_id,
+                    positions: vec![0, 1],
+                },
+                StateEvent::NodeRate {
+                    object_id,
+                    rate: 44100,
+                },
+                StateEvent::NodeVolumes {
+                    object_id,
+                    volumes: vec![0.0, 0.0],
+                },
+                StateEvent::NodeMute {
+                    object_id,
+                    mute: false,
+                },
             ];
             for event in events {
-                state.update(&mut capture_manager, event);
+                state.update(&wirehose, event);
             }
         }
 
-        (state, capture_manager)
+        (state, wirehose)
     }
 
     #[test]
     fn object_list_up_overflow() {
-        let (state, _) = init();
-        let view = View::from(&state, &config::Names::default());
+        let (state, wirehose) = init();
+        let view = View::from(&wirehose, &state, &config::Names::default());
 
         let height = NodeWidget::height() + NodeWidget::spacing();
         // + 2 for header and footer
@@ -570,8 +584,8 @@ mod tests {
 
     #[test]
     fn object_list_down_overflow() {
-        let (state, _) = init();
-        let view = View::from(&state, &config::Names::default());
+        let (state, wirehose) = init();
+        let view = View::from(&wirehose, &state, &config::Names::default());
 
         let height = NodeWidget::height() + NodeWidget::spacing();
         // + 2 for header and footer
