@@ -5,15 +5,12 @@ use std::collections::HashMap;
 
 use serde_json::json;
 
-use crate::command::Command;
 use crate::config;
 use crate::device_kind::DeviceKind;
-use crate::media_class::MediaClass;
-use crate::object::ObjectId;
-use crate::state;
+use crate::wirehose::{media_class, state, CommandSender, ObjectId};
 
-/// A view for transforming [`State`](`crate::state::State`) into a better
-/// format for rendering.
+/// A view for transforming [`State`](`state::State`) into a better format for
+/// rendering.
 ///
 /// This is done in only two ways:
 ///
@@ -26,11 +23,10 @@ use crate::state;
 /// get the initial state from PipeWire. Peak updates happen very frequently
 /// though, hence the optimization.
 ///
-/// There are also functions like [`Self::mute()`] for returning
-/// [`Command`](`crate::command::Command`)s which can be sent to the
-/// [`monitor`](`crate::monitor`).
-#[derive(Debug, Default)]
-pub struct View {
+/// There are also functions like [`Self::mute()`] for executing commands
+/// against [`wirehose`](`crate::wirehose`).
+pub struct View<'a> {
+    wirehose: &'a dyn CommandSender,
     pub nodes: HashMap<ObjectId, Node>,
     pub devices: HashMap<ObjectId, Device>,
 
@@ -51,7 +47,7 @@ pub struct View {
     pub metadata_id: Option<ObjectId>,
 }
 
-#[derive(Debug, Clone, Copy, PartialEq)]
+#[derive(Debug, Clone, Copy, PartialEq, PartialOrd)]
 pub enum Target {
     Node(ObjectId),
     Route(ObjectId, i32, i32),
@@ -61,12 +57,12 @@ pub enum Target {
 
 #[derive(Debug)]
 pub struct Node {
-    pub id: ObjectId,
-    pub object_serial: i32,
+    pub object_id: ObjectId,
+    pub object_serial: u64,
     pub name: String,
     pub title: String,
     pub title_source_sink: Option<String>,
-    pub media_class: MediaClass,
+    pub media_class: String,
 
     pub routes: Option<Vec<(Target, String)>>,
 
@@ -80,9 +76,8 @@ pub struct Node {
     pub positions: Option<Vec<u32>>,
 
     /// If this is a device/endpoint node, store the (device_id, route_index,
-    /// card_device) here because they are needed for the
-    /// [`DeviceVolume`](`crate::command::DeviceVolume`) and
-    /// [`DeviceMute`](`crate::command::DeviceMute`) commands.
+    /// card_device) here because they are needed for changing volumes and
+    /// muting via [`wirehose`](`crate::wirehose`).
     pub device_info: Option<(ObjectId, i32, i32)>,
 
     pub is_default_sink: bool,
@@ -91,8 +86,8 @@ pub struct Node {
 
 #[derive(Debug)]
 pub struct Device {
-    pub id: ObjectId,
-    pub object_serial: i32,
+    pub object_id: ObjectId,
+    pub object_serial: u64,
     pub title: String,
 
     pub profiles: Vec<(Target, String)>,
@@ -140,7 +135,7 @@ impl ListKind {
 /// given media class.
 fn route_targets(
     device: &state::Device,
-    media_class: &MediaClass,
+    media_class: &String,
 ) -> Option<Vec<(Target, String)>> {
     let profile_index = device.profile_index?;
     let profile = device.profiles.get(&profile_index)?;
@@ -166,7 +161,7 @@ fn route_targets(
                     format!("{} (unavailable)", route.description)
                 };
                 Some((
-                    Target::Route(device.id, route.index, *route_device),
+                    Target::Route(device.object_id, route.index, *route_device),
                     title,
                 ))
             })
@@ -199,25 +194,25 @@ impl Node {
         default_source_name: &Option<String>,
         node: &state::Node,
     ) -> Option<Node> {
-        let id = node.id;
+        let object_id = node.object_id;
 
-        let media_class = node.media_class.as_ref()?.clone();
+        let media_class = node.props.media_class()?.clone();
         let title = names.resolve(state, node)?;
 
         // Nodes can represent either streams or devices.
         let (volumes, mute, device_info) =
-            if let Some(device_id) = node.device_id {
+            if let Some(device_id) = node.props.device_id() {
                 // Nodes for devices should get their volume and mute status
                 // from the associated device's active route which is also used
                 // for changing the volume and mute status.
-                let device = state.devices.get(&device_id)?;
-                let card_device = node.card_profile_device?;
+                let device = state.devices.get(device_id)?;
+                let card_device = *node.props.card_profile_device()?;
                 if let Some(route) = active_route(device, card_device) {
                     let route_index = route.index;
                     (
                         route.volumes.clone(),
                         route.mute,
-                        Some((device_id, route_index, card_device)),
+                        Some((*device_id, route_index, card_device)),
                     )
                 } else {
                     (node.volumes.as_ref()?.clone(), node.mute?, None)
@@ -229,11 +224,11 @@ impl Node {
             };
 
         let (routes, target, target_title) = if let Some(device_id) =
-            node.device_id
+            node.props.device_id()
         {
             // Targets for device nodes are routes for the associated device.
-            let device = state.devices.get(&device_id)?;
-            let card_device = node.card_profile_device?;
+            let device = state.devices.get(device_id)?;
+            let card_device = *node.props.card_profile_device()?;
 
             let mut routes: Vec<_> =
                 route_targets(device, &media_class).unwrap_or_default();
@@ -250,7 +245,7 @@ impl Node {
                     };
                     (
                         Some(Target::Route(
-                            device.id,
+                            device.object_id,
                             route.index,
                             card_device,
                         )),
@@ -261,14 +256,14 @@ impl Node {
             };
 
             (Some(routes), target, target_title)
-        } else if media_class.is_sink_input() {
+        } else if media_class::is_sink_input(&media_class) {
             // Targets for output streams are sinks.
-            let outputs = state.outputs(id);
+            let outputs = state.outputs(object_id);
             let sink = sinks.iter().find(|(target, _)| {
                 matches!(target, Target::Node(sink_id)
                     if outputs.contains(sink_id))
             });
-            let (target, target_title) = if !has_target(state, node.id) {
+            let (target, target_title) = if !has_target(state, node.object_id) {
                 (
                     Some(Target::Default),
                     sink.map(|(_, title)| title.clone())
@@ -281,14 +276,14 @@ impl Node {
                 )
             };
             (None, target, target_title)
-        } else if media_class.is_source_output() {
+        } else if media_class::is_source_output(&media_class) {
             // Targets for input streams are sources.
-            let inputs = state.inputs(id);
+            let inputs = state.inputs(object_id);
             let source = sources.iter().find(|(target, _)| {
                 matches!(target, Target::Node(source_id)
                     if inputs.contains(source_id))
             });
-            let (target, target_title) = if !has_target(state, node.id) {
+            let (target, target_title) = if !has_target(state, node.object_id) {
                 (
                     Some(Target::Default),
                     source
@@ -307,11 +302,11 @@ impl Node {
         };
 
         Some(Self {
-            id,
-            object_serial: node.object_serial?,
-            name: node.name.as_ref()?.clone(),
+            object_id,
+            object_serial: *node.props.object_serial()?,
+            name: node.props.node_name()?.clone(),
             title,
-            title_source_sink: node.media_name.clone(),
+            title_source_sink: node.props.media_name().cloned(),
             media_class,
             routes,
             target,
@@ -321,8 +316,10 @@ impl Node {
             peaks: node.peaks.clone(),
             positions: node.positions.clone(),
             device_info,
-            is_default_sink: *default_sink_name == node.name,
-            is_default_source: *default_source_name == node.name,
+            is_default_sink: default_sink_name.as_ref()
+                == node.props.node_name(),
+            is_default_source: default_source_name.as_ref()
+                == node.props.node_name(),
         })
     }
 }
@@ -333,7 +330,7 @@ impl Device {
         device: &state::Device,
         names: &config::Names,
     ) -> Option<Device> {
-        let id = device.id;
+        let object_id = device.object_id;
 
         let title = names.resolve(state, device)?;
 
@@ -352,7 +349,7 @@ impl Device {
         profiles.sort_by_key(|&(index, _)| index);
         let profiles = profiles
             .into_iter()
-            .map(|(index, title)| (Target::Profile(id, index), title))
+            .map(|(index, title)| (Target::Profile(object_id, index), title))
             .collect();
 
         let target_profile = device.profiles.get(&device.profile_index?)?;
@@ -362,11 +359,13 @@ impl Device {
             format!("{} (unavailable)", target_profile.description)
         };
 
-        let target = Some(Target::Profile(id, device.profile_index?));
+        let target = Some(Target::Profile(object_id, device.profile_index?));
+
+        let object_serial = *device.props.object_serial()?;
 
         Some(Device {
-            id,
-            object_serial: device.object_serial?,
+            object_id,
+            object_serial,
             title,
             profiles,
             target_title,
@@ -378,8 +377,8 @@ impl Device {
 fn default_for(state: &state::State, which: &str) -> Option<String> {
     let metadata = state.get_metadata_by_name("default")?;
     let json = metadata.properties.get(&0)?.get(which)?;
-    let obj = serde_json::from_str::<serde_json::Value>(json).ok()?;
-    Some(String::from(obj["name"].as_str()?))
+    let object = serde_json::from_str::<serde_json::Value>(json).ok()?;
+    Some(String::from(object["name"].as_str()?))
 }
 
 fn target_node(state: &state::State, node_id: ObjectId) -> Option<i64> {
@@ -408,9 +407,32 @@ fn has_target(state: &state::State, node_id: ObjectId) -> bool {
     }
 }
 
-impl View {
+impl<'a> View<'a> {
+    pub fn new(wirehose: &'a dyn CommandSender) -> View<'a> {
+        Self {
+            wirehose,
+            nodes: Default::default(),
+            devices: Default::default(),
+            nodes_all: Default::default(),
+            nodes_playback: Default::default(),
+            nodes_recording: Default::default(),
+            nodes_output: Default::default(),
+            nodes_input: Default::default(),
+            devices_all: Default::default(),
+            sinks: Default::default(),
+            sources: Default::default(),
+            default_sink: Default::default(),
+            default_source: Default::default(),
+            metadata_id: Default::default(),
+        }
+    }
+
     /// Create a View from scratch from a provided State.
-    pub fn from(state: &state::State, names: &config::Names) -> View {
+    pub fn from(
+        wirehose: &'a dyn CommandSender,
+        state: &state::State,
+        names: &config::Names,
+    ) -> View<'a> {
         let default_sink_name = default_for(state, "default.audio.sink");
         let default_source_name = default_for(state, "default.audio.source");
 
@@ -419,8 +441,10 @@ impl View {
                 state
                     .nodes
                     .values()
-                    .find(|node| node.name.as_ref() == Some(default_sink_name))
-                    .map(|node| Target::Node(node.id))
+                    .find(|node| {
+                        node.props.node_name() == Some(default_sink_name)
+                    })
+                    .map(|node| Target::Node(node.object_id))
             });
 
         let default_source =
@@ -431,17 +455,20 @@ impl View {
                         .nodes
                         .values()
                         .find(|node| {
-                            node.name.as_ref() == Some(default_source_name)
+                            node.props.node_name() == Some(default_source_name)
                         })
-                        .map(|node| Target::Node(node.id))
+                        .map(|node| Target::Node(node.object_id))
                 });
 
         let mut sinks: Vec<_> = state
             .nodes
             .values()
             .filter_map(|node| {
-                if node.media_class.as_ref()?.is_sink() {
-                    Some((Target::Node(node.id), names.resolve(state, node)?))
+                if media_class::is_sink(node.props.media_class()?) {
+                    Some((
+                        Target::Node(node.object_id),
+                        names.resolve(state, node)?,
+                    ))
                 } else {
                     None
                 }
@@ -454,14 +481,14 @@ impl View {
             .nodes
             .values()
             .filter_map(|node| {
-                if node.media_class.as_ref()?.is_source() {
+                if media_class::is_source(node.props.media_class()?) {
                     let title = names.resolve(state, node)?;
-                    Some((Target::Node(node.id), title))
-                } else if node.media_class.as_ref()?.is_sink() {
+                    Some((Target::Node(node.object_id), title))
+                } else if media_class::is_sink(node.props.media_class()?) {
                     let title = names.resolve(state, node)?;
                     Some((
-                        Target::Node(node.id),
-                        format!("Monitor of {}", title),
+                        Target::Node(node.object_id),
+                        format!("Monitor of {title}"),
                     ))
                 } else {
                     None
@@ -485,14 +512,14 @@ impl View {
                     node,
                 )
             })
-            .map(|node| (node.id, node))
+            .map(|node| (node.object_id, node))
             .collect();
 
         let devices: HashMap<ObjectId, Device> = state
             .devices
             .values()
             .filter_map(|device| Device::from(state, device, names))
-            .map(|device| (device.id, device))
+            .map(|device| (device.object_id, device))
             .collect();
 
         let mut nodes_all = Vec::new();
@@ -504,16 +531,16 @@ impl View {
             nodes.iter().sorted_by_key(|(_, node)| node.object_serial)
         {
             nodes_all.push(*id);
-            if node.media_class.is_sink_input() {
+            if media_class::is_sink_input(&node.media_class) {
                 nodes_playback.push(*id);
             }
-            if node.media_class.is_source_output() {
+            if media_class::is_source_output(&node.media_class) {
                 nodes_recording.push(*id);
             }
-            if node.media_class.is_sink() {
+            if media_class::is_sink(&node.media_class) {
                 nodes_output.push(*id);
             }
-            if node.media_class.is_source() {
+            if media_class::is_source(&node.media_class) {
                 nodes_input.push(*id);
             }
         }
@@ -530,6 +557,7 @@ impl View {
             .collect();
 
         Self {
+            wirehose,
             nodes,
             devices,
             nodes_all,
@@ -549,7 +577,7 @@ impl View {
     /// Update just the peaks of an existing State.
     pub fn update_peaks(&mut self, state: &state::State) {
         for state_node in state.nodes.values() {
-            if let Some(node) = self.nodes.get_mut(&state_node.id) {
+            if let Some(node) = self.nodes.get_mut(&state_node.object_id) {
                 match &state_node.peaks {
                     Some(peaks) => {
                         let peaks_ref =
@@ -563,118 +591,118 @@ impl View {
         }
     }
 
-    /// Returns a command for setting the provided node as the default
-    /// source/sink, depending on device_kind.
-    pub fn set_default(
-        &self,
-        node_id: ObjectId,
-        device_kind: DeviceKind,
-    ) -> Option<Command> {
-        let node = self.nodes.get(&node_id)?;
+    /// Sets the provided node as the default source/sink, depending on
+    /// device_kind.
+    pub fn set_default(&self, node_id: ObjectId, device_kind: DeviceKind) {
+        let Some(node) = self.nodes.get(&node_id) else {
+            return;
+        };
+        let Some(metadata_id) = self.metadata_id else {
+            return;
+        };
+
         let key = match device_kind {
             DeviceKind::Source => "default.configured.audio.source",
             DeviceKind::Sink => "default.configured.audio.sink",
         };
-        let metadata_id = self.metadata_id?;
 
-        Some(Command::MetadataSetProperty(
+        self.wirehose.metadata_set_property(
             metadata_id,
             0,
             String::from(key),
             Some(String::from("Spa:String:JSON")),
             Some(json!({ "name": &node.name }).to_string()),
-        ))
+        );
     }
 
-    /// Returns a command for setting the provided node's target to the
-    /// provided target.
-    pub fn set_target(
-        &self,
-        node_id: ObjectId,
-        target: Target,
-    ) -> Vec<Command> {
+    /// Sets the provided node's target to the provided target.
+    pub fn set_target(&self, node_id: ObjectId, target: Target) {
         let Some(metadata_id) = self.metadata_id else {
-            return Vec::new();
+            return;
         };
 
         match target {
             Target::Default => {
-                vec![
-                    Command::MetadataSetProperty(
-                        metadata_id,
-                        node_id.into(),
-                        String::from("target.object"),
-                        Some(String::from("Spa:Id")),
-                        Some(String::from("-1")),
-                    ),
-                    Command::MetadataSetProperty(
-                        metadata_id,
-                        node_id.into(),
-                        String::from("target.node"),
-                        Some(String::from("Spa:Id")),
-                        Some(String::from("-1")),
-                    ),
-                ]
+                self.wirehose.metadata_set_property(
+                    metadata_id,
+                    node_id.into(),
+                    String::from("target.object"),
+                    Some(String::from("Spa:Id")),
+                    Some(String::from("-1")),
+                );
+                self.wirehose.metadata_set_property(
+                    metadata_id,
+                    node_id.into(),
+                    String::from("target.node"),
+                    Some(String::from("Spa:Id")),
+                    Some(String::from("-1")),
+                );
             }
             Target::Node(target_id) => {
-                vec![
-                    Command::MetadataSetProperty(
-                        metadata_id,
-                        node_id.into(),
-                        String::from("target.object"),
-                        None,
-                        None,
-                    ),
-                    Command::MetadataSetProperty(
-                        metadata_id,
-                        node_id.into(),
-                        String::from("target.node"),
-                        Some(String::from("Spa:Id")),
-                        Some(target_id.to_string()),
-                    ),
-                ]
+                self.wirehose.metadata_set_property(
+                    metadata_id,
+                    node_id.into(),
+                    String::from("target.object"),
+                    None,
+                    None,
+                );
+                self.wirehose.metadata_set_property(
+                    metadata_id,
+                    node_id.into(),
+                    String::from("target.node"),
+                    Some(String::from("Spa:Id")),
+                    Some(target_id.to_string()),
+                );
             }
             Target::Route(device_id, route_index, route_device) => {
-                vec![Command::DeviceSetRoute(
+                self.wirehose.device_set_route(
                     device_id,
                     route_index,
                     route_device,
-                )]
+                );
             }
             Target::Profile(device_id, profile_index) => {
-                vec![Command::DeviceSetProfile(device_id, profile_index)]
+                self.wirehose.device_set_profile(device_id, profile_index);
             }
         }
     }
 
-    /// Returns a command for muting the provided node.
-    pub fn mute(&self, node_id: ObjectId) -> Option<Command> {
-        let node = self.nodes.get(&node_id)?;
+    /// Mutes the provided node.
+    pub fn mute(&self, node_id: ObjectId) {
+        let Some(node) = self.nodes.get(&node_id) else {
+            return;
+        };
+
         let mute = !node.mute;
 
         if let Some((device_id, route_index, route_device)) = node.device_info {
-            Some(Command::DeviceMute(
+            self.wirehose.device_mute(
                 device_id,
                 route_index,
                 route_device,
                 mute,
-            ))
+            );
         } else {
-            Some(Command::NodeMute(node_id, mute))
+            self.wirehose.node_mute(node_id, mute);
         }
     }
 
-    /// Returns a command for changing the volume of the provided node.
+    /// Changes the volume of the provided node. If max volume is provided,
+    /// won't change volume if result would be greater than max. Returns true
+    /// if volume was changed, otherwise false.
     pub fn volume(
         &self,
         node_id: ObjectId,
         adjustment: VolumeAdjustment,
-    ) -> Option<Command> {
-        let node = self.nodes.get(&node_id)?;
+        max: Option<f32>,
+    ) -> bool {
+        let Some(node) = self.nodes.get(&node_id) else {
+            return false;
+        };
 
         let mut volumes = node.volumes.clone();
         if volumes.is_empty() {
-            return None;
+            return false;
         }
         match adjustment {
             VolumeAdjustment::Relative(delta) => {
@@ -687,19 +715,30 @@ impl View {
         }
         let volumes = volumes;
 
+        if let Some(max) = max {
+            if volumes
+                .iter()
+                .any(|volume| (volume.cbrt() * 100.0).round() > max)
+            {
+                return false;
+            }
+        }
+
         if let Some((device_id, route_index, route_device)) = node.device_info {
-            Some(Command::DeviceVolumes(
+            self.wirehose.device_volumes(
                 device_id,
                 route_index,
                 route_device,
                 volumes,
-            ))
+            );
         } else {
-            Some(Command::NodeVolumes(node_id, volumes))
+            self.wirehose.node_volumes(node_id, volumes);
         }
+
+        true
     }
 
-    fn ids(&self, node_kind: ListKind) -> &[ObjectId] {
+    fn object_ids(&self, node_kind: ListKind) -> &[ObjectId] {
         match node_kind {
             ListKind::Node(NodeKind::Playback) => &self.nodes_playback,
             ListKind::Node(NodeKind::Recording) => &self.nodes_recording,
@@ -712,7 +751,7 @@ impl View {
 
     /// Gets all the nodes without filtering.
     pub fn full_nodes(&self, node_kind: NodeKind) -> Vec<&Node> {
-        let node_ids = self.ids(ListKind::Node(node_kind));
+        let node_ids = self.object_ids(ListKind::Node(node_kind));
         node_ids
             .iter()
             .filter_map(|node_id| self.nodes.get(node_id))
@@ -721,7 +760,7 @@ impl View {
 
     /// Gets all the devices without filtering.
     pub fn full_devices(&self) -> Vec<&Device> {
-        let device_ids = self.ids(ListKind::Device);
+        let device_ids = self.object_ids(ListKind::Device);
         device_ids
             .iter()
             .filter_map(|device_id| self.devices.get(device_id))
@@ -734,7 +773,7 @@ impl View {
         list_kind: ListKind,
         object_id: Option<ObjectId>,
     ) -> Option<ObjectId> {
-        let objects = self.ids(list_kind);
+        let objects = self.object_ids(list_kind);
         let next_index = match object_id {
             Some(object_id) => objects
                 .iter()
@@ -751,7 +790,7 @@ impl View {
         list_kind: ListKind,
         object_id: Option<ObjectId>,
     ) -> Option<ObjectId> {
-        let objects = self.ids(list_kind);
+        let objects = self.object_ids(list_kind);
         let next_index = match object_id {
             Some(object_id) => objects
                 .iter()
@@ -768,12 +807,14 @@ impl View {
         list_kind: ListKind,
         object_id: ObjectId,
     ) -> Option<usize> {
-        self.ids(list_kind).iter().position(|&id| id == object_id)
+        self.object_ids(list_kind)
+            .iter()
+            .position(|&id| id == object_id)
     }
 
     /// Returns length of the list_kind.
     pub fn len(&self, list_kind: ListKind) -> usize {
-        self.ids(list_kind).len()
+        self.object_ids(list_kind).len()
     }
 
     /// Returns the possible targets for a node.
@@ -786,9 +827,9 @@ impl View {
         // Get the target list appropriate to the node type
         let (mut targets, default) = if let Some(routes) = &node.routes {
             (routes.clone(), None)
-        } else if node.media_class.is_sink_input() {
+        } else if media_class::is_sink_input(&node.media_class) {
             (self.sinks.clone(), self.default_sink)
-        } else if node.media_class.is_source_output() {
+        } else if media_class::is_source_output(&node.media_class) {
             (self.sources.clone(), self.default_source)
         } else {
             (Vec::new(), None)
@@ -799,14 +840,14 @@ impl View {
                 targets
                     .iter()
                     .find(|(target, _)| *target == default)
-                    .map(|(_, name)| format!("Default: {}", name))
+                    .map(|(_, name)| format!("Default: {name}"))
             })
             .unwrap_or(String::from("Default: No default"));
         // Sort targets by name
         targets.sort_by(|(_, a), (_, b)| a.cmp(b));
         // If the targets are nodes, add the default node to the top
-        if node.media_class.is_sink_input()
-            || node.media_class.is_source_output()
+        if media_class::is_sink_input(&node.media_class)
+            || media_class::is_source_output(&node.media_class)
         {
             targets.insert(0, (Target::Default, default_name.clone()));
         };
